""" Averaging kernel utilities """

from . import cubesphere
from . import ice
from . import latlon
from . import land_lm3
from . import land_lm4
from . import tripolar

__all__ = ["cubesphere", "ice", "land_lm3", "land_lm4", "latlon", "tripolar"]



""" Generic Suite of Utilities """

import math
import pickle
import sqlite3
import warnings

import numpy as np
import pkg_resources as pkgr

__all__ = [
    "get_web_vars_dict",
    "mask_latitude_bands",
    "area_mean",
    "legacy_area_mean",
    "cube_sphere_aggregate",
    "write_sqlite_data",
    "parse_cell_measures",
    "extract_metadata",
    "write_metadata",
]


def get_web_vars_dict():
    """Returns dictionary of legacy LM3 variables

    Returns
    -------
    dict
        LM3 variable module mappings and metadata
    """
    mapping_file = pkgr.resource_filename(
        "gfdlvitals", "resources/LM3_variable_dictionary.pkl"
    )
    return pickle.load(open(mapping_file, "rb",))


def mask_latitude_bands(var, cell_area, geolat, region="global"):
    """Masks a variable and associated cell area based on region

    Parameters
    ----------
    var : np.ma.MaskedArray
        Input array
    cell_area : np.ma.MaskedArray
        Array of cell areas
    geolat : np.ma.MaskedArray
        Array of latitude coordinates
    region : str, optional
        Region of "global", "nh", "sh", or "tropics", by default "global"

    Returns
    -------
    np.ma.MaskedArray
        var, cell_area masked by region

    Raises
    ------
    ValueError
        Unknown region specified
    """
    if region == "tropics":
        var = np.ma.masked_where(np.logical_or(geolat < -30.0, geolat > 30.0), var)
        cell_area = np.ma.masked_where(
            np.logical_or(geolat < -30.0, geolat > 30.0), cell_area
        )
    elif region == "nh":
        var = np.ma.masked_where(np.less_equal(geolat, 30.0), var)
        cell_area = np.ma.masked_where(np.less_equal(geolat, 30.0), cell_area)
    elif region == "sh":
        var = np.ma.masked_where(np.greater_equal(geolat, -30.0), var)
        cell_area = np.ma.masked_where(np.greater_equal(geolat, -30.0), cell_area)
    elif region != "global":
        raise ValueError("Unknown region specified.")
    return var, cell_area


def area_mean(
    var, cell_area, geolat, geolon, region="global", cell_depth=None,
):
    """Computes area mean of a variable

    Parameters
    ----------
    var : np.ma.MaskedArray
        Input array
    cell_area : np.ma.MaskedArray
        Array of cell areas
    geolat : np.ma.MaskedArray
        Array of latitude coordinates
    geolon : np.ma.MaskedArray
        Array of longitude coordinates
    region : str, optional
        Region of "global", "nh", "sh", or "tropics", by default "global"
    cell_depth : np.ma.MaskedArray, optional
        Array of cell depths for 4D-data, by default None

    Returns
    -------
    np.ma.MaskedArray
        Scalar area mean of the variable and cell area
    """
    if cell_depth is not None:
        if var.shape[0] == cell_depth.shape[0]:
            cell_area = np.tile(cell_area[None, :], (cell_depth.shape[0], 1, 1))
            geolat = np.tile(geolat[None, :], (cell_depth.shape[0], 1, 1))
            geolon = np.tile(geolon[None, :], (cell_depth.shape[0], 1, 1))
        else:
            print(
                "Warning: inconsisent dimensions between varname and the cell depth axis.",
                var.shape[0],
                cell_depth.shape[0],
            )
            null_result = np.ma.masked_where(True, 0.0)
            return null_result, null_result
    cell_area = np.ma.array(cell_area)
    cell_area.mask = var.mask
    var, cell_area = mask_latitude_bands(var, cell_area, geolat, region)
    if cell_depth is not None:
        summed = np.ma.sum(
            var
            * cell_area
            * np.tile(cell_depth[:, None, None], (1, var.shape[1], var.shape[2]))
        )
        var = np.ma.average(var, axis=0, weights=cell_depth)
        res = np.ma.sum(var * cell_area) / cell_area.sum()
        return res, summed.sum()
    res = np.ma.sum(var * cell_area) / cell_area.sum()
    return res, cell_area.sum()


def legacy_area_mean(
    var,
    cell_area,
    geolat,
    geolon,
    cell_frac=None,
    soil_frac=None,
    region="global",
    varname=None,
    cell_depth=None,
    component=None,
):
    """Legacy version of area mean calculation

    Parameters
    ----------
    var : np.ma.MaskedArray
        Input array
    cell_area : np.ma.MaskedArray
        Array of cell areas
    geolat : np.ma.MaskedArray
        Array of latitude coordinates
    geolon : np.ma.MaskedArray
        Array of longitude coordinates
    cell_frac : np.ma.MaskedArray, optional
        Array of land cell fraction per grid cell, by default None
    soil_frac : np.ma.MaskedArray, optional
        Array of soil fraction per grid cell, by default None
    region : str, optional
        Region of "global", "nh", "sh", or "tropics", by default "global"
    varname : str, optional
        Variable name, by default None
    cell_depth : np.ma.MaskedArray, optional
        Array of cell depths for 4D-data, by default None
    component : str, optional
        Model component, by default None

    Returns
    -------
    np.ma.MaskedArray
        Scalar area mean of the variable and cell area
    """
    # Land-specific modifications
    if component == "land":
        module_dict = get_web_vars_dict()
        # Read dictionary of keys
        if varname in module_dict.keys():
            module = module_dict[varname]
        elif varname.lower() in module_dict.keys():
            module = module_dict[varname.lower()]
        else:
            module = ""
        # Create a weighting factor
        if module == "vegn":
            cell_area = cell_area * cell_frac * soil_frac
        else:
            cell_area = cell_area * cell_frac
        # Create a 3-D mask if needed
        if cell_depth is not None:
            if var.shape[0] == cell_depth.shape[0]:
                cell_area = np.tile(cell_area[None, :], (cell_depth.shape[0], 1, 1))
                geolat = np.tile(geolat[None, :], (cell_depth.shape[0], 1, 1))
                geolon = np.tile(geolon[None, :], (cell_depth.shape[0], 1, 1))
            else:
                print(
                    "Warning: inconsisent dimensions between varname and the cell depth axis.",
                    var.shape[0],
                    cell_depth.shape[0],
                )
                null_result = np.ma.masked_where(True, 0.0)
                return null_result, null_result
        # Apply data mask to weighting mask
        cell_area.mask = var.mask
    var, cell_area = mask_latitude_bands(var, cell_area, geolat, region)
    # -- Land depth averaging and summation
    if cell_depth is not None:
        summed = np.ma.sum(
            var
            * cell_area
            * np.tile(cell_depth[:, None, None], (1, var.shape[1], var.shape[2]))
        )
        var = np.ma.average(var, axis=0, weights=cell_depth)
        res = np.ma.sum(var * cell_area) / cell_area.sum()
        return res, summed
    res = np.ma.sum(var * cell_area) / cell_area.sum()
    summed = np.ma.sum(var * cell_area)
    return res, summed


def cube_sphere_aggregate(var, tiles):
    """[summary]

    Parameters
    ----------
    var : str
        Variable name
    tiles : netCDF4.Dataset
        Datasets for each cubesphere tile

    Returns
    -------
    np.ma.MaskedArray
        Concatenated array of data across of cubesphere faces
    """

    assert len(tiles) == 6, "Six cubesphere tiles must be provided."

    return np.ma.concatenate(
        (
            tiles[0].variables[var][:],
            tiles[1].variables[var][:],
            tiles[2].variables[var][:],
            tiles[3].variables[var][:],
            tiles[4].variables[var][:],
            tiles[5].variables[var][:],
        ),
        axis=-1,
    )


def write_sqlite_data(
    sqlfile, varname, fyear, varmean=None, varsum=None, component=None
):
    """Writes data to sqlite file

    Parameters
    ----------
    sqlfile : str, path-like
        Path to output sqlite file
    varname : str
        Variable name
    fyear : str
        Year being processed
    varmean : float, optional
        Mean of the data, by default None
    varsum : float, optional
        Sum of variable of the data, by default None
    component : str, optional
        Model component, by default None
    """

    missing_value = -1.0e20

    # check if result is a nan and replace with a defined missing value
    if varmean is not None:
        if math.isnan(float(varmean)):
            warnings.warn(
                f"Could not update {sqlfile} variable {varname} with mean={varmean}"
            )
            varmean = missing_value

    if varsum is not None:
        if math.isnan(float(varsum)):
            warnings.warn(
                f"Could not update {sqlfile} variable {varname} with mean={varsum}"
            )
            varsum = missing_value

    conn = sqlite3.connect(sqlfile)
    cur = conn.cursor()
    if component == "land":
        sql = (
            "create table if not exists "
            + varname
            + " (year integer primary key, sum float, avg float)"
        )
    else:
        sql = (
            "create table if not exists "
            + varname
            + " (year integer primary key, value float)"
        )
    cur.execute(sql)
    if component == "land":
        sql = (
            "insert or replace into "
            + varname
            + " values("
            + fyear[:4]
            + ","
            + str(varsum)
            + ","
            + str(varmean)
            + ")"
        )
    else:
        sql = (
            "insert or replace into "
            + varname
            + " values("
            + fyear[:4]
            + ","
            + str(varmean)
            + ")"
        )
    cur.execute(sql)
    conn.commit()
    cur.close()
    conn.close()


def parse_cell_measures(attr, key):
    """Parse cell measures attribute

    Parameters
    ----------
    attr : str
        Cell measurues attribute string
    key : str
        Variable name key

    Returns
    -------
    str
        Returns truncated cell measures string
    """
    if attr is not None:
        ind = attr.split().index(key + ":") + 1
        return attr.split()[ind]
    return None


def extract_metadata(dset, varname, attr):
    """Obtain a specific variable attribute

    Parameters
    ----------
    dset : netCDF4.Dataset
        Input NetCDF dataset
    varname : str
        Variable name
    attr : str
        Requested attribute

    Returns
    -------
    str
        Variable attribute string
    """
    if attr in dset.variables[varname].__dict__.keys():
        return dset.variables[varname].__dict__[attr]
    return None


def write_metadata(sqlfile, varname, attr, value):
    """Writes dataset metadata to sqlite file

    Parameters
    ----------
    sqlfile : str, path-like
        Path to output sqlite file
    varname : str
        Variable name
    attr : str
        Attribute name
    value : str
        Attribute string
    """
    if value is None:
        value = str("")
    conn = sqlite3.connect(sqlfile)
    cur = conn.cursor()
    sql = (
        "create table if not exists "
        + str(attr)
        + " (var text primary key, value text)"
    )
    cur.execute(sql)
    sql = (
        "insert or replace into "
        + str(attr)
        + ' values("'
        + str(varname)
        + '","'
        + str(value)
        + '")'
    )
    cur.execute(sql)
    conn.commit()
    cur.close()
    conn.close()


def standard_grid_cell_area(lat, lon, earth_radius=6371.0e3):
    """Calculate grid cell area for a standard grid

    Parameters
    ----------
    lat : numpy.ndarray
        Array of latitude axis
    lon : numpy.ndarray
        Array of longitude axis
    earth_radius : float, optional
        Radius of the Earth, by default 6371.0e3

    Returns
    -------
    numpy.ndarray
        Array of cell areas
    """
    dlat = lat[1] - lat[0]
    dlon = lon[1] - lon[0]
    area = np.empty((len(lat), len(lon)))
    for j, _lat in enumerate(lat):
        for i, _lon in enumerate(lon):
            lon1 = _lon + dlon / 2.0
            lon0 = _lon - dlon / 2.0
            lat1 = _lat + dlat / 2.0
            lat0 = _lat - dlat / 2.0
            area[j, i] = (
                (np.pi / 180.0)
                * earth_radius
                * earth_radius
                * np.abs(np.sin(np.radians(lat0)) - np.sin(np.radians(lat1)))
                * np.abs(lon0 - lon1)
            )
    return area



""" Module for parsing FMS timings """

import tarfile as tf
import io
import pandas as pd
from gfdlvitals.util import gmeantools


__all__ = ["timing"]


def timing(ascii_file, fyear, outdir, label):
    """Extracts FMS timings

    Parameters
    ----------
    ascii_file : str, path-like
        Path to ascii tar file
    fyear : str
        Year to process (YYYY)
    outdir : str
        Path for output SQLite file
    label : str
        Name of output SQLite file
    """

    def ascii_tar_to_stats_df(ascii_file):
        """Subroutine to read fms.out from ascii tar file,
        extract the clock timings, and return a pandas dataframe

        Parameters
        ----------
        ascii_file : str, path-like
            Path to ascii tar file

        Returns
        -------
        pandas.DataFrame of timings
        """

        member = None
        tar = tf.open(ascii_file)
        for member in tar.getnames():
            if "fms.out" in member:
                break

        txtfile = tar.extractfile(member)
        content = txtfile.readlines()
        x = -1
        for x, line in enumerate(content):
            if "Total runtime" in line.decode("utf-8"):
                break
        tar.close()
        content = content[x::]

        output = io.BytesIO()
        output.write(str.encode("clock,min,max,mean,std\n"))
        for line in content:
            line = line.decode("utf-8")
            if not line.startswith(" "):
                line = line.split()
                line = line[-1::-1]
                label = " ".join(line[8::][-1::-1])
                label = label.replace(" ", "_")
                label = label.replace("-", "_")
                label = label.replace("&", "_")
                for x in ["(", ")", "*", "/", ":"]:
                    label = label.replace(x, "")
                line = [label] + line[0:8][-1::-1][0:4]
                line = ",".join(line) + "\n"
                output.write(str.encode(line))
        output.seek(0)

        df = pd.read_csv(output, delimiter=",")
        # df = df.set_index('clock')

        return df

    df = ascii_tar_to_stats_df(ascii_file)

    clocks = sorted(df["clock"].to_list())
    for clock in clocks:
        for attr in ["mean", "min", "max"]:
            val = df[df["clock"] == clock][attr].values[0]

            # -- Write to sqlite
            gmeantools.write_sqlite_data(
                outdir + "/" + fyear + ".globalAve" + label + ".db",
                clock + "_" + attr,
                fyear[:4],
                val,
            )



""" Legacy land model averaging routines """

import multiprocessing

from functools import partial

import numpy as np

from gfdlvitals.util.average import RichVariable
from gfdlvitals.util.average import process_var

import gfdlvitals.util.netcdf as nctools

__all__ = ["average"]


def average(grid_file, data_file, fyear, out, lab):
    """Mid-level averaging routine

    Parameters
    ----------
    grid_file : bytes-like
        Gridspec dataset
    data_file : bytes-like
        Model output dataset
    fyear : str
        Year being processed
    out : str
        Output path directory
    lab : [type]
        DB file name
    """

    _grid_file = nctools.in_mem_nc(grid_file)
    _data_file = nctools.in_mem_nc(data_file)

    lat = _data_file["lat"][:]
    lon = _data_file["lon"][:]
    geolon, geolat = np.meshgrid(lon, lat)

    # land areas and fractions
    # global cell_area

    cell_area = _grid_file["land_area"][:]
    cell_frac = _grid_file["land_frac"][:]

    if "soil_area" in _grid_file.variables.keys():
        soil_area = _grid_file["soil_area"][0]
    elif "soil_area" in _data_file.variables.keys():
        soil_area = _data_file["soil_area"][0]
    else:
        raise ValueError("Unable to locate soil area field.")
    soil_frac = np.ma.array(soil_area / (cell_area * cell_frac))

    variables = list(_data_file.variables.keys())
    variables = [
        RichVariable(
            x,
            grid_file,
            data_file,
            fyear,
            out,
            lab,
            geolat,
            geolon,
            cell_area=cell_area,
            cell_frac=cell_frac,
            soil_area=soil_area,
            soil_frac=soil_frac,
        )
        for x in variables
    ]

    pool = multiprocessing.Pool(multiprocessing.cpu_count())
    pool.map(partial(process_var, **{"averager": "land-lm3"}), variables)



""" Routine for calculating AMOC """

import warnings

try:
    import xoverturning
except:
    warnings.warn("Could not load xoverturning.")
import numpy as np
import xarray as xr
from gfdlvitals.util import gmeantools
from gfdlvitals.util.netcdf import tar_member_exists
from gfdlvitals.util.netcdf import extract_from_tar
from gfdlvitals.util.netcdf import in_mem_xr


__all__ = ["mom6_amoc"]


def mom6_amoc(fyear, tar, label="Ocean", outdir="./"):
    """Driver for AMOC calculation in MOM6-class models

    Parameters
    ----------
    fyear : str
        Year label (YYYY)
    tar : tarfile
        In-memory history tarfile object
    label : str
        SQLite output stream name
    outdir : str, path-like
        Path to output SQLite file
    """

    member = f"{fyear}.ocean_annual_z.nc"
    static = f"{fyear}.ocean_static.nc"

    annual_file = (
        extract_from_tar(tar, member, ncfile=True)
        if tar_member_exists(tar, member)
        else None
    )
    static_file = (
        extract_from_tar(tar, static, ncfile=True)
        if tar_member_exists(tar, static)
        else None
    )

    if annual_file is not None and static_file is not None:
        # open the Dataset with the transports
        dset = in_mem_xr(annual_file)

        # select first time level from static file
        # editorial comment: why does the static file have a time dimension?
        dset_static = in_mem_xr(static_file).isel(time=0)

        # merge static DataSet with transport DataSet
        for geo_coord in ["geolon_v", "geolat_v", "wet_v"]:
            if geo_coord in dset_static.variables:
                dset[geo_coord] = xr.DataArray(
                    dset_static[geo_coord].values, dims=dset_static[geo_coord].dims
                )

        required_vars = ["geolon_v", "geolat_v", "umo", "vmo"]
        dset_vars = list(dset.variables)

        if list(set(required_vars) - set(dset_vars)) == []:
            # calculate non-rotated y-ward moc array
            moc = xoverturning.calcmoc(dset, basin="atl-arc", verbose=False)

            # max streamfunction between 20N-80N and 500-2500m depth
            maxsfn = moc.sel(yq=slice(20.0, 80.0), z_i=slice(500.0, 2500.0)).max()
            maxsfn = maxsfn.astype(np.float16).values
            print(f"  AMOC = {maxsfn}")

            # max streamfunction at 26.5N
            rapidsfn = moc.sel(yq=26.5, method="nearest")
            rapidsfn = rapidsfn.sel(z_i=slice(500.0, 2500.0)).max()
            rapidsfn = rapidsfn.astype(np.float16).values
            print(f"  RAPID AMOC = {rapidsfn}")

            # -- Write to sqlite
            gmeantools.write_sqlite_data(
                outdir + "/" + fyear + ".globalAve" + label + ".db",
                "amoc_vh",
                fyear[:4],
                maxsfn,
            )

            gmeantools.write_sqlite_data(
                outdir + "/" + fyear + ".globalAve" + label + ".db",
                "amoc_rapid",
                fyear[:4],
                rapidsfn,
            )

        else:
            warnings.warn(f"{required_vars} are required to calculate AMOC")

    else:
        warnings.warn("AMOC calculation requires ocean_static and ocean_annual_z")



""" Standardized plotting routines """

import warnings
import cftime
import nc_time_axis

import numpy as np
import pkg_resources as pkgr

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import font_manager

import gfdlvitals

__all__ = ["set_font", "plot_timeseries", "update_figure", "on_key", "run_plotdb"]

COUNT = 1


def set_font():
    """Sets font style to Roboto"""
    # Add Roboto font
    fonts_dir = pkgr.resource_filename("gfdlvitals", "resources/fonts")

    font_dirs = [fonts_dir]
    font_files = font_manager.findSystemFonts(fontpaths=font_dirs)

    for font_file in font_files:
        font_manager.fontManager.addfont(font_file)

    # Define fonts and sizes
    matplotlib.rcParams["font.family"] = "Roboto"
    matplotlib.rcParams.update({"font.size": 14})


def plot_timeseries(
    dsets,
    var,
    trend=False,
    align_times=False,
    plottype="average",
    smooth=None,
    nyears=None,
    labels=None,
    legend=True,
    means=True,
):
    """Standardized function to make a timeseries plot

    Parameters
    ----------
    dsets : gfdlvitals.VitalsDataFrame or list
        Dataframe or list of dataframes to plot
    var : str
        Variable name to plot
    trend : bool, optional
        Plot linear trend line if True, by default False
    align_times : bool, optional
        H, by default False
    plottype : str, optional
        Defines plot type as either "average" or "sum",
        by default "average"
    smooth : int, optional
        Integer number of years to apply smoothing, by default None
    nyears : int, optional
        Limit the x-axis to nyears number of points, by default None
    labels : str, optional
        Comma-separated list of dataset labels, by default None
    legend : bool, optional
        Display a legend for the plot, by default True
    means : bool, optional
        Add variable means to the legend, by default True

    Returns
    -------
    matplotlib.pyplot.figure, dict
        Matplotlib figure handle and dictionary of axes/dataset mappings
    """

    set_font()

    # Ensure "dsets" is a list
    dsets = [dsets] if not isinstance(dsets, list) else dsets

    # Text Labels
    if labels is None:
        labels = [f"Dataset {x}" for x in range(0, len(dsets))]
        legend = False
    else:
        labels = labels.split(",")

    # Determine max length of time values
    maxlen = max([len(x.index) for x in dsets])

    # summed
    if plottype == "sum":
        dsets = [x.areasum() for x in dsets]

    # Compute trends if asked
    if trend:
        trends = [x.trend()[0:nyears] for x in dsets]

    if align_times:
        dsets = [x.extend(maxlen) for x in dsets]

    # Setup the figure. The top half will have a 16:9 aspect
    # ratio. The bottom half will be used for legend info and
    # the whole figure will be cropped at the end

    fig = plt.figure(figsize=(12, 6.75))
    ax1 = plt.subplot(1, 1, 1)

    # Establish a list of axes in the figure
    axes_dict = {}

    # If smoothing is requested, still plot a faint copy of the full timeseries
    for x, dset in enumerate(dsets):
        dset.attrs["alpha"] = 0.3 if smooth is not None else 1.0
        dset.attrs["color"] = f"C{x}"

    _lines = []

    for i, dset in enumerate(dsets):
        label = labels[i]

        axes_dict[label] = {}
        axes_dict[label]["data"] = dset

        # Determine if we need a twin time axis
        _ax = ax1.twiny() if align_times and i > 0 else ax1
        axes_dict[label]["axis"] = _ax

        axes_list = [axes_dict[x]["axis"] for x in list(axes_dict.keys())]

        # Keep a list of the axes and move new ones to the bottom of the figure
        if _ax not in list(axes_dict.keys()) and align_times:
            # Move twinned axis ticks and label from top to bottom
            _ax.xaxis.set_ticks_position("bottom")
            _ax.xaxis.set_label_position("bottom")

            # Offset the twin axis below the host
            _ax.spines["bottom"].set_position(("axes", -0.10 * (len(axes_dict) - 1)))

            # Make sure frame is displayed
            if i > 0:
                _ax.set_frame_on(True)
                # Turn on the frame for the twin axis, but then hide all
                # but the bottom spine
                _ax.patch.set_visible(False)
                _ = [sp.set_visible(False) for sp in _ax.spines.values()]
                _ax.spines["bottom"].set_visible(True)

        # Convert times to nc_time_axis
        times = [
            nc_time_axis.CalendarDateTime(item, "noleap") for item in dset.index.values
        ]


        # Add means to the labels
        _label = labels[i]
        if means:
            _mean = float(np.nanmean(dset[var].values[0:nyears]))
            _mean = round(_mean, 4)
            _label = f"{_label}  (mean={_mean})"
            labels[i] = _label

        # Make the first plot
        (axes_dict[label]["line"],) = _ax.plot(
            times[0:nyears],
            dset[var].values[0:nyears],
            color=dset.attrs["color"],
            alpha=dset.attrs["alpha"],
            label=_label,
        )

        _lines.append(axes_dict[label]["line"])

        # After the first plot is established, align time axes if asked
        if align_times:
            _time_index = dset.index[0:nyears]
            _ax.set_xlim(
                cftime.date2num(
                    _time_index[0], calendar="noleap", units="days since 2000-01-01"
                ),
                cftime.date2num(
                    _time_index[-1], calendar="noleap", units="days since 2000-01-01"
                ),
            )

        if trend:
            (axes_dict[label]["trendline"],) = _ax.plot(
                times[0 : len(trends[i])],
                trends[i][var].values[0:nyears],
                linestyle="dashed",
                color=dset.attrs["color"],
                alpha=1.0,
                linewidth=1,
            )
            axes_dict[label]["trend"] = trends[i]

        if smooth:
            (axes_dict[label]["smoothline"],) = _ax.plot(
                times[0:nyears],
                dset.smooth(smooth)[var].values[0:nyears],
                color=dset.attrs["color"],
                alpha=1.0,
                linewidth=2,
            )

        # Text annotations

        if i == 0:
            axes_dict[label]["topline_label"] = _ax.text(
                0.01, 1.08, var, ha="left", transform=ax1.transAxes, fontsize=22
            )

            _long_name = dset[var].attrs["long_name"]
            _long_name = f"Sum of {_long_name}" if plottype == "sum" else _long_name


            axes_dict[label]["longname_label"] = _ax.text(
                0.01,
                1.03,
                _long_name,
                ha="left",
                transform=ax1.transAxes,
                style="italic",
                fontsize=14,
                fontfamily="Roboto Condensed",
            )

            axes_dict[label]["units_label"] = _ax.set_ylabel(dset[var].attrs["units"])

    axes_list = [axes_dict[x]["axis"] for x in list(axes_dict.keys())]
    axes_list = list(set(axes_list))
    maxlim = max([x.get_xlim()[1] - x.get_xlim()[0] for x in axes_list])
    _ = [x.set_xlim(x.get_xlim()[0], x.get_xlim()[0] + maxlim) for x in axes_list]

    # Add grid
    _ = [x.grid(color="gray", linestyle="--", linewidth=0.3) for x in axes_list]

    if legend:
        plt.legend(
            _lines,
            labels,
            fancybox=True,
            framealpha=1,
            shadow=True,
            borderpad=0.5,
            loc="upper center",
            bbox_to_anchor=(0.5, (-0.10 * (len(axes_dict)) + 0.02)),
        )

    return fig, axes_dict


def update_figure(fig, axes_dict, varname, smooth, nyears, trend):
    """Update plot on keypress of left and right arrow keys

    Parameters
    ----------
    fig : matplotlib.pyplot.figure
        Matplotlib figure handle
    axes_dict : dict
        Internal structure of axes associations to the data
    varname : str
        Variable name to plot
    smooth : int, None type
        Integer number of years to apply as a smoother
    nyears : int, None type
        Integer number to years to limit the plotting
    trend : bool
        Add linear trend lines if True, omit if False

    Returns
    -------
    matplotlib.pyplot.figure
        Matplotlib figure handle
    """
    for i, label in enumerate(list(axes_dict.keys())):
        axes_dict[label]["line"].set_ydata(
            axes_dict[label]["data"][varname].values[0:nyears]
        )

        if trend:
            axes_dict[label]["trendline"].set_ydata(
                axes_dict[label]["trend"][varname].values[0:nyears]
            )

        if smooth:
            axes_dict[label]["smoothline"].set_ydata(
                axes_dict[label]["data"].smooth(smooth)[varname].values[0:nyears]
            )

        if i == 0:
            axes_dict[label]["topline_label"].set_text(varname)
            axes_dict[label]["longname_label"].set_text(
                axes_dict[label]["data"][varname].attrs["long_name"]
            )
            axes_dict[label]["units_label"].set_text(
                axes_dict[label]["data"][varname].attrs["units"]
            )

    axes_list = [axes_dict[x]["axis"] for x in list(axes_dict.keys())]
    _ = [ax.relim() for ax in axes_list]
    _ = [ax.autoscale(enable=True, axis="y") for ax in axes_list]

    fig.canvas.draw()

    return fig


def on_key(event, fig, varlist, axes_dict, smooth, nyears, trend):
    """Update plot on keypress of left and right arrow keys

    Parameters
    ----------
    event : mpl keyboard event
        Key press
    fig : matplotlib.pyplot.figure
        Matplotlib figure handle
    varlist : list
        List of variables to plot
    axes_dict : dict
        Internal structure of axes associations to the data
    smooth : int, None type
        Integer number of years to apply as a smoother
    nyears : int, None type
        Integer number to years to limit the plotting
    trend : bool
        Add linear trend lines if True, omit if False
    """
    global COUNT

    redraw = False

    if (event.key == "right") or (event.key == "left"):
        COUNT = (COUNT + 1) if event.key == "right" else (COUNT - 1)
        COUNT = 0 if COUNT < 0 else COUNT
        COUNT = len(varlist) - 1 if COUNT >= len(varlist) else COUNT
        varname = varlist[COUNT]
        redraw = True

    if redraw is True:
        _ = gfdlvitals.plot.update_figure(
            fig, axes_dict, varname, smooth, nyears, trend
        )


def run_plotdb(cliargs):
    """Intermediate function to execute plotting routines

    Parameters
    ----------
    cliargs : argparse.arguments
        Processed command-line arguments
    """
    dsets = [gfdlvitals.open_db(x) for x in cliargs.dbfiles]

    all_variables = [set(x.columns) for x in dsets]
    common_variables = set.intersection(*all_variables)

    extra_variables = [sorted(list(x - common_variables)) for x in all_variables]

    for x, _varlist in enumerate(extra_variables):
        if len(_varlist) > 0:
            print(
                "The following variables are availble only "
                + f"{cliargs.dbfiles[x]} and will not be plotted:\n"
            )

            if len(_varlist) % 2 != 0:
                _varlist.append(" ")

            split = int(len(_varlist) / 2)
            _list1 = _varlist[0:split]
            _list2 = _varlist[split:]
            for key, value in zip(_list1, _list2):
                print("{0:<20s} {1}".format(key, value))
            print("\n")

    variable_list = sorted(list(common_variables))

    mplfig, axes = gfdlvitals.plot.plot_timeseries(
        dsets,
        variable_list[0],
        trend=cliargs.trend,
        align_times=cliargs.align,
        smooth=cliargs.smooth,
        nyears=cliargs.nyears,
        labels=cliargs.labels,
    )
    _ = mplfig.canvas.mpl_connect(
        "key_press_event",
        lambda event: on_key(
            event,
            mplfig,
            variable_list,
            axes,
            cliargs.smooth,
            cliargs.nyears,
            cliargs.trend,
        ),
    )

    plt.tight_layout()
    plt.show(block=True)



""" Utilities for working with NetCDF files """

import tarfile
import netCDF4
import xarray as xr

__all__ = ["extract_from_tar", "in_mem_nc", "in_mem_xr", "tar_member_exists"]


def extract_from_tar(tar, member, ncfile=False):
    """Extract individual file from a tar file

    Parameters
    ----------
    tar : either tarfile object or path str
        Path to tar file
    member : str
        Filename to extract from tar file
    ncfile : bool, optional
        Convert to netCDF4.Dataset object, by default False

    Returns
    -------
    in-memory object
        either byte stram or netCDF4.Dataset
    """

    _tar = tarfile.open(tar) if isinstance(tar, str) else tar

    if member not in _tar.getnames():
        member = "./" + member

    data = _tar.extractfile(member)

    if ncfile:
        data = in_mem_nc(data)

    if isinstance(tar, str):
        _tar.close()

    return data


def in_mem_nc(data):
    """Wrapper to convert bytes object to netCDF4.Dataset

    Parameters
    ----------
    data : byte stream object
        In-memory object

    Returns
    -------
    netCDF4.Dataset
        In-memory netCDF4 dataset object
    """

    data = data.read()
    return netCDF4.Dataset("in-mem-file", mode="r", memory=data)


def in_mem_xr(data):
    """Wrapper to convert bytes object to xarray.Dataset

    Parameters
    ----------
    data : byte stream object
        In-memory object

    Returns
    -------
    xarray.Dataset
        In-memory xarray dataset object
    """

    if isinstance(data, netCDF4._netCDF4.Dataset):
        dfile = xr.open_dataset(xr.backends.NetCDF4DataStore(data), use_cftime=True)
    else:
        dfile = xr.open_dataset(data, use_cftime=True)

    return dfile


def tar_member_exists(tar, member):
    """Tests if file exsits inside a tar file

    Parameters
    ----------
    tar : tarfile object
        Opened tarfile handle
    member : str
        Name of file inside tar file

    Returns
    -------
    bool
        True if exists, otherwise False
    """
    if member in tar.getnames():
        status = True
    elif str("./" + member) in tar.getnames():
        status = True
    else:
        status = False

    return status



""" Utilities for merging DB files """

import sqlite3

__all__ = ["merge"]


def merge(source, destination):
    """Merges two sqlite files

    Parameters
    ----------
    source : str, path-like
        Path to source sqlite file
    destination : str, path-like
        Path to destination sqlite file
    """
    con = sqlite3.connect(destination)
    cur = con.cursor()
    sql = "ATTACH '" + source + "' as src"
    cur.execute(sql)
    cur.close()
    cur = con.cursor()
    sql = "SELECT * FROM main.sqlite_master WHERE type='table'"
    cur.execute(sql)
    cur.close()
    cur = con.cursor()
    sql = "SELECT * FROM src.sqlite_master WHERE type='table'"
    cur.execute(sql)
    src_tables = cur.fetchall()
    cur.close()
    for var in src_tables:
        varname = var[1]
        if varname not in [x[1] for x in src_tables]:
            cur = con.cursor()
            cur.execute(var[-1])
            cur.close()
        cur = con.cursor()
        sql = "INSERT OR REPLACE into " + varname + " SELECT * FROM src." + varname
        cur.execute(sql)
        cur.close()
    con.commit()
    con.close()



""" Pandas class extension for gfdlvitals """

import copy
import math
import sqlite3
import warnings

import cftime

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from scipy import stats

__all__ = [
    "VitalsDataFrame",
    "Timeseries",
    "open_db",
    "reformat_time_axis",
    "ttest_ind_auto",
]


def _autocorr(arr, lag=1):
    """Computes the sample autocorrelation function coeffficient (rho)
    for given lag

    Parameters
    ----------
    arr : numpy.array (1-d)
        Input data array
    lag : int, optional
        lag, by default 1

    Returns
    -------
    float
        lagged auto-correlation
    """

    flen = float(len(arr))
    ybar = float(sum(arr)) / flen
    denom = sum([(y - ybar) ** 2 for y in arr])
    numer = sum(
        [(y - ybar) * (ytpk - ybar) for (y, ytpk) in zip(arr[:-lag], arr[lag:])]
    )
    return numer / denom


def ttest_ind_auto(arr1, arr2, axis=0):
    """Performs a t-test that adjusts the degrees of freedom
    based on the autocorrelation of the dataset.
    See Krasting et al. 2013 for more details
    DOI: 10.1175/JCLI-D-12-00832.1

    Parameters
    ----------
    arr1 : numpy.ndarray
        First array of data points
    arr2 : numpy.ndarray
        Second array of data points
    axis : int, optional
        Axis to perform t-test, by default 0

    Returns
    -------
    tuple
        t-statistic, probability,
        arr1 autocorrelation, arr2 autocorrelation
    """
    arr1, arr2, axis = _chk2_asarray(arr1, arr2, axis)
    variance1 = np.var(arr1, axis, ddof=1)
    variance2 = np.var(arr2, axis, ddof=1)
    arrlen1 = arr1.shape[axis]
    arrlen2 = arr2.shape[axis]
    lag1r1 = _autocorr(np.split(arr1, arr1.shape[axis], axis))
    lag1r2 = _autocorr(np.split(arr2, arr2.shape[axis], axis))
    n1eff = arrlen1 * ((1 - lag1r1) / (1 + lag1r1))
    n2eff = arrlen2 * ((1 - lag1r2) / (1 + lag1r2))
    df = n1eff + n2eff - 2
    diff = np.mean(arr1, axis) - np.mean(arr2, axis)
    svar = ((arrlen1 - 1) * variance1 + (arrlen2 - 1) * variance2) / float(df)
    t = diff / np.sqrt(svar * (1.0 / arrlen1 + 1.0 / arrlen2))
    t = np.where((diff == 0) * (svar == 0), 1.0, t)  # define t=0/0 = 0, identical means
    prob = stats.distributions.t.sf(np.abs(t), df) * 2
    # use np.abs to get upper tail
    # distributions.t.sf currently does not propagate nans
    # this can be dropped, if distributions.t.sf propagates nans
    # if this is removed, then prob = prob[()] needs to be removed
    prob = np.where(np.isnan(t), np.nan, prob)
    if t.ndim == 0:
        t = t[()]
        prob = prob[()]
    result = [t, prob, lag1r1, lag1r2]
    result = [float(x) for x in result]
    return tuple(result)


def _chk2_asarray(arr1, arr2, axis):
    if axis is None:
        arr1 = np.ravel(arr1)
        arr2 = np.ravel(arr2)
        outaxis = 0
    else:
        arr1 = np.asarray(arr1)
        arr2 = np.asarray(arr2)
        outaxis = axis
    return arr1, arr2, outaxis


def _remove_trend(x, y, order=1, anomaly=True, return_coefs=False, coefs=None):
    """Internal function to remove a linear trend

    Parameters
    ----------
    x : numpy.ndarray
        independent axis
    y : np.ndarray
        dependent axis
    order : int, optional
        Polynomial order to use for fitting, by default 1
    anomaly : bool, optional
        Results as anomalies from the fit, by default True
    return_coefs : bool, optional
        Return polynomial fit coeffiecients, by default False
    coefs : tuple, optional
        Use a provided set of coefficients, by default None

    Returns
    -------
    numpy.ndarray
        Detrended data
    """

    if None in list(y):
        return None
    if coefs is None:
        idx = np.isfinite(x) & np.isfinite(y)
        coefs = np.polyfit(x[idx], y[idx], order)
        if return_coefs is True:
            return coefs
    model = np.poly1d(coefs)
    fit = model(x)
    if anomaly is True:
        result = y - fit
    else:
        fit = fit - fit[0]
        result = y - fit
    return result


def _calc_trend(x, y, order=1):
    """Internal function to calculate trend line/curves

    Parameters
    ----------
    x : numpy.ndarray
        independent axis
    y : np.ndarray
        dependent axis
    order : int, optional
        Polynomial order to use for fitting, by default 1

    Returns
    -------
    numpy.ndarray
        Polynomial fit
    """

    idx = np.isfinite(x) & np.isfinite(y)
    coefs = np.polyfit(x[idx], y[idx], order)
    model = np.poly1d(coefs)
    return model(x)


def _remove_reference_trend(t, x, other, anomaly=True):
    """Removes trends from a reference dataset

    Parameters
    ----------
    t : numpy.ndarray
        independent axis
    x : gfdlvitals.VitalsDataFrame
        Dataset to be detrended
    other : gfdlvitals.VitalsDataFrame
        Reference dataset containing the trend to be removed
    anomaly : bool, optional
        Results as anomalies from the fit, by default True

    Returns
    -------
    gfdlvitals.VitalsDataFrame
        Detrended dataset
    """

    if str(x.name) not in list(other.columns):
        result = None
    else:
        _coefs = other[x.name].to_numpy()
        result = _remove_trend(t, x, anomaly=anomaly, coefs=_coefs)
    return result


def reformat_time_axis(ax=None):
    """Reformats x-axis labels to YYYY format

    Parameters
    ----------
    ax : matplotlob.pyplot.figure.axis, optional
        axis object to be reformatted, by default None
    """

    if ax is None:
        ax = plt.gca()
    labels = [x.get_text() for x in ax.xaxis.get_ticklabels()]
    labels = [x.split("-")[0] for x in labels]
    _ = ax.set_xticklabels(labels)


class VitalsDataFrame(pd.DataFrame):
    """Pandas class extension for holding the Vitals data

    Parameters
    ----------
    pd : pandas.DataFrame
        Input Pandas DataFrame object

    Returns
    -------
    gfdlvitals.VitalsDataFrame
        Class extension object
    """

    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ["internal_cache"]
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ["added_property"]

    @property
    def _constructor(self):
        return VitalsDataFrame

    def build_netrad_toa(self):
        """Constructs netrad_toa from component terms if available

        Parameters
        ----------
        self : VitalsDataFrame

        Returns
        -------
        VitalsDataFrame
        """

        varlist = list(self.columns)
        if ("netrad_toa" not in varlist) and (
            all(x in varlist for x in ["swdn_toa", "swup_toa", "olr"])
        ):
            self["netrad_toa"] = self["swdn_toa"] - self["swup_toa"] - self["olr"]

            self["netrad_toa"].attrs = {
                "long_name": "Net TOA Radiation "
                + "(resconstructed as swdn_toa - swup_toa - olr)",
                "units": "W m-2",
            }

        return self

    def areasum(self):
        """Returns the area integrated variable based
        on the cell_measure method
        """
        cell_measures = [(x, self[x].attrs["cell_measure"]) for x in list(self.columns)]
        varlist, cm_set = zip(*cell_measures)
        cm_set = [x for x in list(set(cm_set)) if x is not None]
        attrs = {x: self[x].attrs for x in varlist}
        cell_measures = [x for x in cell_measures if x[0] not in cm_set]

        result = {}

        for x in cell_measures:
            if x[1] is not None:
                try:
                    _res = self[x[0]] * self[x[1]]
                    _res.attrs = attrs[x[0]]
                    result[x[0]] = _res

                except Exception as exc:
                    warnings.warn(f"Unable to sum field {x[0]}")

        result = pd.DataFrame(result)
        result = VitalsDataFrame(result)

        for var in list(result.columns):
            result[var].attrs = attrs[var]

        return result

    def ttest(self, df2):
        """Performs t-test between two instances of VitalsDataFrame

        Parameters
        ----------
        df2 : VitalsDataFrame
            Comparison data set

        Returns
        -------
        pandas.DataFrame
            Contains p-values for variables common between the
            two VitalsDataFrames
        """

        # get a list of common variables
        varlist = list(set(self.columns).intersection(df2.columns))

        # ignore some known fields
        ignore_list = ["area"]
        varlist = [x for x in varlist if x not in ignore_list]

        # perform t-test
        result = {k: ttest_ind_auto(self[k], df2[k])[1] for k in varlist}

        return pd.DataFrame(result, index=["pval"]).transpose()

    def smooth(self, window, extrap=False):
        """Apply a smoother to the dataset

        Parameters
        ----------
        window : int
            Smoothing filter length
        extrap : bool, optional
            Extrapolate data on the ends, by default False

        Returns
        -------
        self
            Smoothed dataset
        """

        if window is None:
            _df = self
        else:
            _df = self.rolling(window, center=True).mean()
            if extrap is True:
                _df.fillna(method="ffill", inplace=True)
                _df.fillna(method="bfill", inplace=True)
        _df.attrs = self.attrs
        return _df

    def extend(self, maxlen):
        """Extend VitalsDataFrame to a set length and pad with NaNs

        Parameters
        ----------
        maxlen : int
            New length of the VitalsDataFrame

        Returns
        -------
        self
            Extended dataset
        """
        endyear = tuple(self.index[-1].timetuple())
        padding = np.arange(1, maxlen - len(self.index) + 1) + endyear[0]
        added_index = [cftime.DatetimeNoLeap(x, *endyear[1:]) for x in padding]
        _df = pd.DataFrame({"times": added_index}).set_index("times")
        _df = self.append(_df)
        _df.attrs = self.attrs
        for column in self.columns:
            _df[column].attrs = self[column].attrs
        return _df

    def detrend(self, reference=None, order=1, anomaly=True, return_coefs=False):
        """Detrend VitalsDataFrame object

        Parameters
        ----------
        reference : gfdlvitals.DataFrame, optional
            Reference VitalsDataFrame, by default None
        order : int, optional
            Polynomial order to use for fitting, by default 1
        anomaly : bool, optional
            Results as anomalies from the fit, by default True
        return_coefs : bool, optional
            Return polynomial fit coeffiecients, by default False

        Returns
        -------
        self
            Extended dataset
        """
        tindex = np.array(
            [
                cftime.date2num(x, "days since 0001-01-01", calendar="noleap")
                for x in self.index
            ]
        )
        if reference is not None:
            if order != 1:
                print(
                    "Only a linear trend can be removed based on "
                    + "another dataset. Setting order to 1."
                )
                order = 1
            coefs = reference.detrend(order=order, return_coefs=True)
            result = self.apply(
                lambda x: (_remove_reference_trend(tindex, x, coefs, anomaly=anomaly))
            )
        else:
            result = self.apply(
                lambda x: (
                    _remove_trend(
                        tindex,
                        x,
                        order=order,
                        anomaly=anomaly,
                        return_coefs=return_coefs,
                    )
                )
            )
        result.attrs = self.attrs
        return result

    def trend(self, order=1):
        """Fits a trend to the VitalsDataFrame object

        Parameters
        ----------
        order : int, optional
            Polynomial order to use for fitting, by default 1

        Returns
        -------
        self
            Fitted trend dataset
        """
        tindex = np.array(
            [
                cftime.date2num(x, "days since 0001-01-01", calendar="noleap")
                for x in self.index
            ]
        )
        result = self.apply(lambda x: (_calc_trend(tindex, x, order=order)))
        result.attrs = self.attrs
        return result


class Timeseries:
    """Timeseries class object

    Parameters
    ----------
    f : str, path-like
        Input SQLite file
    var : str
        Variable to extract
    scale : float, optional
        Scale data by this factor, by default 1.0
    multiply_by_area : bool, optional
        Multiply variable by cell area before returning, by default False
    legacy_land : bool, optional
        Read legacy version of the land SQLite files, by default False
    start : int, optional
        Specify start year, by default None
    end : int, optional
        Specify end year, by default None
    """

    def __init__(
        self,
        f,
        var,
        scale=1.0,
        multiply_by_area=False,
        legacy_land=False,
        start=None,
        end=None,
    ):

        # open the sqlite connection
        con = sqlite3.connect(f)
        cur = con.cursor()

        # get the results from the sqlite file for the requested variable
        if legacy_land is True:
            _ = cur.execute("SELECT year,sum FROM " + var)
        else:
            _ = cur.execute("SELECT year,value FROM " + var)
        results = cur.fetchall()

        # split list of tuples into two different lists
        self._t, self._data = zip(*results)
        self._t = np.array(self._t)

        _ = cur.execute("SELECT name FROM sqlite_master where TYPE='table'")
        tables = [str(record[0]) for record in cur.fetchall()]

        if multiply_by_area is True:
            if "cell_measure" in tables:
                _ = cur.execute(
                    "SELECT value FROM cell_measure where var='" + var + "'"
                )
                cell_measure = cur.fetchone()[0]
            else:
                cell_measure = "area"
            _ = cur.execute("SELECT value FROM " + cell_measure)
            area = np.array(cur.fetchall()).squeeze()
            scale = area * scale

        self._data = np.array(self._data) * scale

        if "long_name" in tables:
            _ = cur.execute(f"SELECT value FROM long_name where var='{var}'")
            result = cur.fetchone()
            self.long_name = result[0] if isinstance(result, tuple) else None
        else:
            self.long_name = None

        if "units" in tables:
            _ = cur.execute(f"SELECT value FROM units where var='{var}'")
            result = cur.fetchone()
            self.units = result[0] if isinstance(result, tuple) else None
        else:
            self.units = None

        if "cell_measure" in tables:
            _ = cur.execute(f"SELECT value FROM cell_measure where var='{var}'")
            result = cur.fetchone()
            self.cell_measure = result[0] if isinstance(result, tuple) else None
        else:
            self.cell_measure = None

        # close the connection
        cur.close()
        con.close()

        # filter based on start year and end year
        if start is not None:
            idx = [i for i, val in enumerate(self._t) if val >= start]
            self._t = self._t[idx]
            self._data = self._data[idx]
        else:
            start = self._t.min()
        if end is not None:
            idx = [i for i, val in enumerate(self._t) if val <= end]
            self._t = self._t[idx]
            self._data = self._data[idx]
        else:
            end = self._t.max() + 1

        # check for missing values and pad with nans
        missing_times = set(np.arange(start, end)) - set(self._t)
        if len(list(missing_times)) != 0:
            warnings.warn(f"Timeseries is incomplete for {var}: {missing_times}")

        # pad missing values with nans
        self.dict = dict(zip(self._t, self._data))
        self.dict = {
            **self.dict,
            **dict(zip(missing_times, [np.nan] * len(missing_times))),
        }

    @property
    def t(self):
        k, v = zip(*sorted(self.dict.items()))
        return k

    @property
    def data(self):
        k, v = zip(*sorted(self.dict.items()))
        return v

    def __str__(self):
        return self.__class__.__name__

    def __hash__(self):
        return hash([self.__dict__[x] for x in list(self.__dict__.keys())])


def open_db(
    dbfile,
    variables=None,
    yearshift=0.0,
    legacy_land=False,
    start=None,
    end=None,
):
    """Function to read sqlite dbfile"""

    if variables is None:
        conn = sqlite3.connect(dbfile)
        cur = conn.cursor()
        sql = "SELECT name FROM sqlite_master WHERE type='table'"
        _ = cur.execute(sql)
        variables = [str(record[0]) for record in cur.fetchall()]
        cur.close()
        conn.close()
        removes = ["units", "long_name", "cell_measure"]
        variables = [x for x in variables if x not in removes]

    # -- Loop over variables
    data = {}
    attributes = {}
    years = []
    skipped = []
    for var in variables:
        tsobj = Timeseries(dbfile, var, legacy_land=legacy_land, start=start, end=end)
        if len(tsobj.t) > 0:
            data[var] = tsobj.data
            years = years + list(tsobj.t)
            attributes[var] = {
                "long_name": tsobj.long_name,
                "units": tsobj.units,
                "cell_measure": tsobj.cell_measure,
            }

    years = list(set(years))
    years = [x + float(yearshift) for x in years]

    variables = list(set(variables) - set(skipped))

    if start is None:
        start = -1 * math.inf

    if end is None:
        end = math.inf

    df = pd.DataFrame(data, index=years)
    df = df[(df.index >= start) & (df.index <= end)]
    df.index = cftime.num2date(
        (df.index * 365.0) - (365.0 / 2.0) - 1,
        "days since 0001-01-01",
        calendar="365_day",
    )

    df = VitalsDataFrame(df)
    df = df.sort_index()

    for var in list(df.columns):
        df[var].attrs = attributes[var]

    return df
